#include <getopt.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/capability.h>
#include <sys/prctl.h>
#include <sys/types.h>
#include <unistd.h>

#include "shared/log.h"
#include "shared/util.h"
#include "crash-manager.h"

#define USER_NAME "crash_worker"

static void print_help(const char *name)
{
	printf("Syntax: %s [OPTIONS]\n"
	       "\n"
	       "    -p  --pid=PID           PID of dumped process\n"
	       "    -u  --uid=UID           real UID of dumped process\n"
	       "    -g  --gid=GID           real GID of dumped process\n"
	       "    -i  --tid=TID           TID of thread that triggered core dump\n"
	       "    -s  --signal=SIG        number of signal causing dump\n"
	       "    -t  --time=TIME         time of dump, expressed as seconds since the Epoch\n"
	       "    -l  --live              get coredump of running process\n"
	       "    -k  --kill-after-dump   kill after dump (only with --live option)\n"
	       "    -r  --print             print report path to stdout\n"
	       "    -o  --output            output directory\n"
	       "    -E  --executable-path   path to executable that crashed\n"
	       "    -h  --help              this message\n"
	       "\n"
	       "for --live option only --pid is required\n"
	       "\n", name);
}

static bool parse_args(struct crash_info *cinfo, int argc, char *argv[])
{
#define QUOTE(member) #member
#define GET_NUMBER(member) \
	do {\
		errno = 0; \
		cinfo->member##_info = strtol(optarg, NULL, 10); \
		if (errno != 0) { \
			_D("%s argument error\n", QUOTE(member)); \
			printf("%s argument error\n", QUOTE(member)); \
			return false; \
		} \
	} while(0)

	bool result = true;
	int opt;
	bool pid_set = false;
	bool uid_set = false;
	bool gid_set = false;
	bool sig_set = false;

	struct option long_options[] = {
		{"pid",             required_argument, NULL, 'p'},
		{"uid",             required_argument, NULL, 'u'},
		{"gid",             required_argument, NULL, 'g'},
		{"tid",             required_argument, NULL, 'i'},
		{"signal",          required_argument, NULL, 's'},
		{"time",            required_argument, NULL, 't'},
		{"live",                  no_argument, NULL, 'l'},
		{"kill-after-dump",       no_argument, NULL, 'k'},
		{"print",                 no_argument, NULL, 'r'},
		{"output",          required_argument, NULL, 'o'},
		{"help",                  no_argument, NULL, 'h'},
		{"executable-path", required_argument, NULL, 'E'},
	};

	while ((opt = getopt_long(argc, argv, "p:u:g:i:s:t:hlkro:E:", long_options, NULL)) != -1) {
		switch (opt) {
		case 'p':
			GET_NUMBER(pid);
			pid_set = true;
			break;
		case 'u':
			GET_NUMBER(uid);
			uid_set = true;
			break;
		case 'g':
			GET_NUMBER(gid);
			gid_set = true;
			break;
		case 'i':
			GET_NUMBER(tid);
			break;
		case 's':
			GET_NUMBER(sig);
			sig_set = true;
			break;
		case 't':
			GET_NUMBER(time);
			break;
		case 'l':
			cinfo->livedump = true;
			break;
		case 'k':
			cinfo->kill = true;
			break;
		case 'r':
			cinfo->print_result_path = true;
			break;
		case 'o':
			cinfo->output_path = optarg;
			_D("output path: %s\n", optarg);
			break;
		case 'E':
			cinfo->executable_path = strdup(optarg);
			if (!cinfo->executable_path)
				return false;

			kernel_exe_path_normalize(cinfo->executable_path);
			_D("executable path: %s\n", cinfo->executable_path);
			break;
		case 'h':
		default:
			print_help(argv[0]);
			return false;
		}
	}

	if (cinfo->livedump && !have_livecoredumper()) {
		printf("livecoredumper not available - can not perform livedump.\n");
		return false;
	}

	if (!pid_set || (!cinfo->livedump && (!gid_set || !uid_set || !sig_set))) {
		printf("Not enough parameters.\n\n");
		print_help(argv[0]);
		return false;
	}

	if (cinfo->livedump && sig_set) {
		printf("--sig can not be used with --live option\n\n");
		print_help(argv[0]);
		return false;
	}

	if (!cinfo->livedump && cinfo->kill) {
		printf("Option --kill-after-dump can be used only with --live\n");
		return false;
	}
	return result;
#undef QUOTE
#undef GET_NUMBER
}

static bool set_caps(const cap_flag_t flag)
{
	bool res = false;
	cap_t caps;
	cap_value_t cap_list[5] = {
		CAP_DAC_READ_SEARCH,
		CAP_DAC_OVERRIDE,
		CAP_KILL,
		CAP_SYS_PTRACE,
		CAP_SYSLOG
	};

	caps = cap_get_proc();
	if (caps == NULL) {
		_E("cap_get_proc() error: %m");
		goto exit;
	}

	if (cap_set_flag(caps, flag, ARRAY_SIZE(cap_list), cap_list, CAP_SET) == -1) {
		_E("cap_set_flag() error: %m");
		goto exit;
	}

	if (cap_set_proc(caps) == -1) {
		_E("cap_set_proc() error: %m");
		goto exit;
	}

	res = true;
exit:
	if (caps != NULL && cap_free(caps) == -1) {
		_E("cap_free() error: %m");
		res = false;
	}
	return res;
}

static bool set_groups(const char *user_name, const gid_t gid)
{
	int ngroup = 0;
	errno = 0;
	int ret = getgrouplist(user_name, gid, NULL, &ngroup);
	if (ret != -1) {
		_E("getgrouplist() unexpected return value: %d", ret);
		return false;
	} else if (errno != 0) {
		_E("getgrouplist() unexcepted failure: %m");
		return false;
	}

	gid_t groups[ngroup];
	ret = getgrouplist(user_name, gid, groups, &ngroup);
	if (ret == -1) {
		_E("getgrouplist() error");
		return false;
	}

	ret = setgroups(ngroup, groups);
	if (ret != 0) {
		_E("setgroups() error: %m\n");
		return false;
	}

	return true;
}

static bool drop_privileges(const char *user_name)
{
	struct passwd *user_info = getpwnam(user_name);
	if (!user_info) {
		_E("getpwnam() error: %m");
		return false;
	}

	if (!set_caps(CAP_PERMITTED))
		return false;

	/*
	 * setuid() clears capabilities, so we need to set PR_SET_KEEPCAPS and
	 * restore them after that
	 */
	if (prctl(PR_SET_KEEPCAPS, 1, 0, 0, 0) == -1) {
		_E("prctl(PR_SET_KEEPCAPS) error: %m");
		return false;
	}

	if (!set_groups(user_name, user_info->pw_gid))
		return false;

	if (setuid(user_info->pw_uid) == -1) {
		_E("setuid() error: %m\n");
		return false;
	}

	if (!set_caps(CAP_EFFECTIVE))
		return false;

	if (!set_caps(CAP_INHERITABLE))
		return false;

	return true;
}

int main(int argc, char *argv[])
{
	int res;

	if (!drop_privileges(USER_NAME)) {
		res = EXIT_FAILURE;
		goto exit;
	}

	struct crash_info cinfo;

	/*
	 * prctl(PR_SET_DUMPABLE, 0) is not neccessary. Kernel runs the
	 * crash-manager and sets RLIMIT_CORE to 1 for the process. This is special
	 * value that prevents from running crash-manager recursively.
	 */

	crash_info_init(&cinfo);

	/* Parse args */
	if (!parse_args(&cinfo, argc, argv))
		return EXIT_FAILURE;

	res = crash_manager_direct(&cinfo) ? EXIT_SUCCESS : EXIT_FAILURE;

	crash_manager_free(&cinfo);
exit:
	_I("Exiting with exit code %d", res);
	return res;
}
