/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef LIVEDUMPER_HPP__
#define LIVEDUMPER_HPP__

#include "core.hpp"
#include "helpers.hpp"
#include "prochandle.hpp"
#include "log.hpp"
#include "maps.hpp"

#include <dirent.h>
#include <limits.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ptrace.h>
#include <sys/wait.h>

#include <fstream>
#include <iostream>
#include <stdexcept>
#include <string>
#include <vector>
#include <map>

template <typename T>
class LiveDumper {
 private:
	pid_t m_pid;
	std::vector<pid_t> m_tpids;
	std::map<pid_t,user_regs_struct> m_registers;
	Core<T> m_core;
	std::ofstream m_core_file;
	int prstatus_fd = -1;
	std::pair <void *, size_t> map_buff;

 public:
	explicit LiveDumper(const pid_t pid) : m_pid(pid), map_buff(MAP_FAILED, 0) {}

	~ LiveDumper () {
		if (map_buff.first != MAP_FAILED)
			munmap(map_buff.first, map_buff.second);
	}

	void CollectTpids() {
		m_tpids.clear();

		std::string task_dir = "/proc/" + std::to_string(m_pid) + "/task";

		DIR *dir = opendir(task_dir.c_str());
		if (dir == nullptr)
			throw std::system_error(errno, std::system_category(), "failed to open directory: " + task_dir);

		dirent *ent;
		while ((ent = readdir(dir)) != nullptr) {
			if (ent->d_type == DT_DIR &&
				strncmp(ent->d_name, ".", 1) != 0)
					m_tpids.push_back(std::strtol(ent->d_name, 0, 10));
		}
		closedir(dir);
	}

	void AttachTo(const pid_t pid) const {
		if (ptrace(PTRACE_ATTACH, pid, nullptr, nullptr) == -1)
			throw std::system_error(errno, std::system_category(), "failed to attach to: " + std::to_string(pid));
	}

	void setPrstatusFd(const int prstatus_fd) {
		this->prstatus_fd = prstatus_fd;
	}

	void Attach() {
		AttachTo(m_pid);

		waitpid(m_pid, nullptr, 0);

		CollectTpids();

		for (const auto &pid : m_tpids) {
			if (pid == m_pid) continue;
			AttachTo(pid);
		}
	}

	void Detach() const {
		for (const auto &pid : m_tpids) {
			if (ptrace(PTRACE_DETACH, pid, nullptr, nullptr) == -1) {
				logger.log_error("Detach from PID %d error: %s",
						 pid, std::system_category().default_error_condition(errno).message());
			}
		}
	}

	void GetRegs(user_regs_struct *registers, pid_t pid) const {
#if defined(__aarch64__)
		struct iovec iov = { registers, sizeof *registers };
		if (ptrace(PTRACE_GETREGSET, pid, NT_PRSTATUS, &iov) == -1)
#else
		if (ptrace(PTRACE_GETREGS, pid, nullptr, registers) == -1)
#endif
			throw std::system_error(errno, std::system_category(),
									"failed to get registers for: " + std::to_string(pid));
	}

	std::vector<char> GetAuxv() {
		auto auxv_name = "/proc/" + std::to_string(m_pid) + "/auxv";
		std::ifstream file;

		file.open(auxv_name, std::ios::binary);
		if (!file)
			throw std::system_error(errno, std::system_category(),
									"failed to open auxv file " + auxv_name);

		std::vector<char> auxv(std::istreambuf_iterator<char>{file}, {});
		return auxv;
	}

	void OpenCoredump(const std::string &path) {
		m_core_file.open(path, std::ios::binary);
		if (!m_core_file)
			throw std::system_error(errno, std::system_category(),
									"failed to open coredump file " + path);
	}

	void CloseCoredump() {
		m_core_file.close();
	}

	const std::vector<std::shared_ptr<MapRecord>> ReadMaps() {
		Maps maps(m_pid);
		maps.Parse();
		return maps.Records();
	}

	void SaveRegsToPrStatusFd(const user_regs_struct &registers) {
		prstatus_t prstatus;

		memset(&prstatus, 0, sizeof(prstatus));
		memcpy(&prstatus.pr_reg, &registers, sizeof(prstatus.pr_reg));
		prstatus.pr_pid = m_pid;

		map_buff.first = mmap(nullptr, sizeof(prstatus),
							 PROT_READ | PROT_WRITE,
							 MAP_SHARED,
							 prstatus_fd,
							 0);
		if (map_buff.first == MAP_FAILED)
			return;
		map_buff.second = sizeof prstatus;

		memcpy(static_cast <char *> (map_buff.first), &prstatus, sizeof(prstatus));
	}

	void AddNotes(const std::vector<std::shared_ptr<MapRecord>> &records) {
		user_regs_struct registers;
		GetRegs(&registers, m_pid);

		if (prstatus_fd >= 0)
			SaveRegsToPrStatusFd(registers);

		m_registers[m_pid] = registers;
		m_core.AddPRSTATUSNote(registers, m_pid);

		for (const auto &pid : m_tpids) {
			if (pid == m_pid)
				continue;
			GetRegs(&registers, pid);

			m_registers[pid] = registers;
			m_core.AddPRSTATUSNote(registers, pid);
		}

		m_core.AddFILENote(records);
		std::vector<char> auxv = GetAuxv();
		m_core.AddAUXVNote(auxv);
	}

	void MarkImportantRegions(const std::vector<std::shared_ptr<MapRecord>> &records) {
	}

	std::string GetExePath(const pid_t pid) {
		char BUFF[PATH_MAX];
		std::string exe_path = std::string("/proc/" + std::to_string(pid) + "/exe");
		ssize_t path_len;
		if ((path_len = readlink(exe_path.c_str(), BUFF, sizeof(BUFF))) == -1) {
			std::cout << "GetExePath error from" << exe_path << std::endl;
			throw std::system_error(errno, std::system_category(), "GetExePath() for " + exe_path + " failed");
		}
		return std::string(BUFF, path_len);
	}

	bool DumpCore(const std::string &path, bool minicore) {
		bool result = true;

		try {
			Guardian guard_open_coredump([this, &path]() { OpenCoredump(path);},
					 [this]() { CloseCoredump(); });
			Guardian guard_attach([this]() { Attach(); },
					 [this]() { Detach(); });

			const std::vector<std::shared_ptr<MapRecord>> records = ReadMaps();

			AddNotes(records);
			MarkImportantRegions(records);
			m_core.ImportMapRecords(records);
			m_core.SaveELFHeader(m_core_file);
			m_core.SaveProgramHeadersTable(m_core_file);
			m_core.SaveNotes(m_core_file);
			std::string mem_path = std::string("/proc/" + std::to_string(m_pid) + "/mem");
			int mem_fd = open(mem_path.c_str(), O_RDONLY);
			if (mem_fd == -1)
				throw std::system_error(errno, std::system_category(), "open() for " + mem_path + " failed");
			m_core.SaveLoadable(mem_fd, m_core_file, minicore);
			if (minicore) {
				m_core.SaveAUXVData(GetExePath(m_pid), mem_fd, m_core_file);
				m_core.SavePthreadList(mem_fd, m_core_file, m_pid);
				m_core.SaveStacks(mem_fd, m_core_file, m_tpids, m_registers);
			}
			close(mem_fd);
		} catch (std::system_error &e) {
			logger.log_error("DumpCore: %s", e.what());
			result = false;
		}

		return result;
	}
};

#endif  // LIVEDUMPER_HPP__
